from .object_base import ObjectBase


class ExternalDocumentation(ObjectBase):
    """
    An `External Documentation Object`_ references external resources for extended
    documentation.

    .. _External Documentation Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#externalDocumentationObject
    """

    __slos__ = ["description", "url"]
    required_fields = "url"

    def _parse_data(self):
        self.description = self._get("description", str)
        self.url = self._get("url", str)


class Reference(ObjectBase):
    """
    A `Reference Object`_ designates a reference to another node in the specification.

    .. _Reference Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#referenceObject
    """

    # can't start a variable name with a $
    __slots__ = ["ref"]
    required_fields = ["$ref"]

    def _parse_data(self):
        self.ref = self._get("$ref", str)

    @classmethod
    def can_parse(cls, dct):
        """
        Override ObjectBase.can_parse because we had to remove the $ from $ref
        in __slots__ (since that's not a valid python variable name)
        """
        # TODO - can a reference object have spec extensions?
        cleaned_keys = [k for k in dct.keys() if not k.startswith("x-")]

        return len(cleaned_keys) == 1 and "$ref" in dct



import sys

from .errors import SpecError, ReferenceResolutionError

IS_PYTHON_2 = False
if sys.version_info[0] == 2:
    IS_PYTHON_2 = True
else:
    # unicode was removed in python3, but we need to support both here, so define
    # it in python 3 only
    unicode = str


def _asdict(x):
    if hasattr(x, "__getstate__"):
        return x.__getstate__()
    elif isinstance(x, dict):
        return {k: _asdict(v) for k, v in x.items()}
    elif isinstance(x, (list, tuple, set)):
        return x.__class__(_asdict(y) for y in x)
    else:
        return x


def raise_on_unknown_type(parent, field, object_types, found):
    """
    Raises a SpecError describing a situation where an unknown type was given.
    This function attempts to produce as useful an error as possible based on the
    type of types that were expected.

    :param parent: The parent element who was attempting to parse the field
    :type parent: Subclass of ObjectBase
    :param field: The field we were trying to parse
    :type field: str
    :param object_types: The types allowed for this field
    :type object_types: List of str or Class
    :param found: The value that was found (and did not match any expected type)
    :type found: any

    :raises: A SpecError describing the failure
    """
    if len(object_types) == 1:
        if isinstance(object_types[0], str):
            expected_type = ObjectBase.get_object_type(object_types[0])
            raise SpecError(
                "Expected {}.{} to be of type {}, with required fields {}".format(
                    parent.get_path(),
                    field,
                    object_types[0],
                    expected_type.required_fields,
                ),
                path=parent.path,
                element=parent,
            )
    elif (
        len(object_types) == 2
        and len([c for c in object_types if isinstance(c, str)]) == 2
        and "Reference" in object_types
    ):
        # we can give a similar error here as above
        expected_type_str = [c for c in object_types if c != "Reference"][0]
        expected_type = ObjectBase.get_object_type(expected_type_str)
        raise SpecError(
            "Expected {}.{} to be of type {} or Reference, but did not find required fields {} or '$ref'".format(
                parent.get_path(),
                field,
                expected_type_str,
                expected_type.required_fields,
            ),
            path=parent.path,
            element=parent,
        )
    print(object_types)
    raise SpecError(
        "Expected {}.{} to be one of [{}], got {}".format(
            parent.get_path(),
            field,
            ",".join(
                [str(c) for c in object_types],
            ),
            type(found),
        ),
        path=parent.path,
        element=parent,
    )


class ObjectBase(object):
    """
    The base class for all schema objects.  Includes helpers for common schema-
    related functions.
    """

    __slots__ = ["path", "raw_element", "_accessed_members", "strict", "_root", "extensions", "_original_ref"]
    required_fields = []

    def __init__(self, path, raw_element, root):
        """
        Creates a new Object for a OpenAPI schema with a reference to its own
        path in the schema.

        :param path: The path to this element in the spec.
        :type path: list[str]
        :param raw_element: The raw element parsed from the spec that this object
                            is parsing.
        :type raw_element: dict
        :param root: The root of the spec, for reference
        :type root: OpenAPI
        """
        # init empty slots
        for k in type(self).__slots__:
            if k in ("_spec_errors", "validation_mode"):
                # allow these two fields to keep their values
                continue
            setattr(self, k, None)

        self.path = path
        self.raw_element = raw_element
        self._root = root

        self._accessed_members = []
        self.extensions = {}

        # TODO - add strict mode that errors if all members were not accessed
        self.strict = False

        # parse our own element
        try:
            self._required_fields(*type(self).required_fields)
            self._parse_data()
        except SpecError as e:
            if self._root.validation_mode:
                self._root.log_spec_error(e)
            else:
                raise

        # TODO - this may not be appropriate in all cases
        self._parse_spec_extensions()

        # TODO - assert that all keys of raw_element were accessed

    def __repr__(self):
        """
        Returns a string representation of the parsed object
        """
        # TODO - why?
        return "<{} {}>".format(type(self), self.path)

    def __getstate__(self):
        """
        Returns this object as a dict, removing all empty keys.  This can be used
        to serialize a spec.

        Allows pickling objects by returning a dict of all slotted values.
        """
        return _asdict({k: getattr(self, k) for k in type(self).__slots__ if hasattr(self, k)})

    def __setstate__(self, state):
        """
        Allows unpickling objects
        """
        for k, v in state.items():
            setattr(self, k, v)

    def _required_fields(self, *fields):
        """
        Given a list of require fields for this object, raises a SpecError if any
        of the fields do not exist.

        :param *fields: A list of fields to ensure exist in this object
        :type *fields: str

        :raises SpecError: if any of the required fields are not present.
        """
        missing_fields = []
        for field in fields:
            if field not in self.raw_element:
                missing_fields.append(field)

        if missing_fields:
            raise SpecError(
                "Missing required fields: {}".format(", ".join(missing_fields)), path=self.path, element=self
            )

    def _parse_data(self):
        """
        Parses the raw_element into this object.  This is not implemented here,
        but is called in the constructor and _must_ be implemented in all
        subclasses.

        An implementation of this method should use :any:`_get` to retrieve
        values from the raw_element, which has the side-effect of noting that
        those members were accessed.  After this is executed, spec extensions
        are parsed and then an assertion is made that all keys in the
        raw_element were accessed - if not, the schema is considered invalid.
        """
        raise NotImplementedError("You must implement this method in subclasses!")

    def _get(self, field, object_types, is_list=False, is_map=False):
        """
        Retrieves a value from this object's raw element, and returns None if
        it is not present.  Use :any:`_required_fields` to ensure all required
        fields are present before depending on the output of this method.

        :param field: The field name to retrieve
        :type field: str
        :param object_types: The types of Objects that are accepted.  One of
                             these types will be returned, or the spec will be
                             considered invalid.  If the magic string '*' is
                             passed in, it must be the only accepted type, and
                             all types will be accepted.
        :type object_types: list[str or Type]
        :param is_list: If true, this should return a List of object of the give
                        types.
        :param is_list: bool
        :param is_map: If true, this must return a :any:`Map` of object of the given
                       types
        :type is_map: bool

        :returns: object_type if given, otherwise the type parsed from the spec
                  file
        """
        self._accessed_members.append(field)

        ret = self.raw_element.get(field, None)
        if ret is None:
            return None

        try:
            if not isinstance(object_types, list):
                # maybe don't accept not-lists
                object_types = [object_types]

            if "*" in object_types and len(object_types) != 1:
                raise ValueError("Fields that accept any type must not specify any other types!")

            # if yaml loads a value that includes a unicode character in python2,
            # that value will come in as a ``unicode`` type instead of a ``str``.
            # For the purposes of this library, those are the same thing, so in
            # python2 only, we'll include ``unicode`` for any element that
            # accepts ``str`` types.
            if IS_PYTHON_2:
                if str in object_types:
                    object_types += [unicode]

            if is_list:
                if not isinstance(ret, list):
                    raise SpecError(
                        "Expected {}.{} to be a list of [{}], got {}".format(
                            self.get_path, field, ",".join([str(c) for c in object_types]), type(ret)
                        ),
                        path=self.path,
                        element=self,
                    )
                ret = self.parse_list(ret, object_types, field)
            elif is_map:
                if not isinstance(ret, dict):
                    raise SpecError(
                        "Expected {}.{} to be a Map of string: [{}], got {}".format(
                            self.get_path, field, ",".join([str(c) for c in object_types]), type(ret)
                        ),
                        path=self.path,
                        element=self,
                    )
                ret = Map(self.path + [field], ret, object_types, self._root)
            else:
                accepts_string = str in object_types
                found_type = False

                for t in object_types:
                    if t == "*":
                        found_type = True
                        break

                    if t == str:
                        # try to parse everything else first
                        continue

                    if isinstance(t, str):
                        # we were given the name of a subclass of ObjectBase,
                        # attempt to parse ret as that type
                        python_type = ObjectBase.get_object_type(t)

                        if python_type.can_parse(ret):
                            ret = python_type(self.path + [field], ret, self._root)
                            found_type = True
                            break
                    elif isinstance(ret, t):
                        # it's already the type we need
                        found_type = True
                        break

                if not found_type:
                    if accepts_string and isinstance(ret, str):
                        found_type = True

                if not found_type:
                    raise_on_unknown_type(self, field, object_types, ret)
        except SpecError as e:
            if self._root.validation_mode:
                self._root.log_spec_error(e)
                ret = None
            else:
                raise

        return ret

    @classmethod
    def key_contained(cls, key, target_list):
        """
        Returns whether the key is contained in the given list or not.
        We use this specific function as to prevent usage of keywords we add "_"
        to several parameters, and we still want to validate those parameters
        """
        if key.endswith("_"):
            extra_key = key[:-1]
        else:
            extra_key = key + "_"
        return key in target_list or extra_key in target_list

    @classmethod
    def can_parse(cls, dct):
        """
        Returns True if this class can parse the given dict.  This is based on
        the __slots__ and required_fields of the class, and the keys of the dict.
        This is intended to be used when an element may be one of a number of
        allowed Object types - each type should independently consider if it
        can parse the given element, and the first to report that it can should
        be used.

        :param dct: The dict to consider.
        :type dct: dict

        :returns: True if this class can parse dct into an instance of itself,
                  otherwise False
        :rtype: bool
        """
        # if this isn't a dict, the spec is dreadfully wrong (and since no type
        # will be able to parse this value, an appropriate error is returned)
        if not isinstance(dct, dict):
            return False
        # ensure that the dict's keys are valid in our slots
        for key in dct.keys():
            if key.startswith("x-"):
                # ignore spec extensions
                continue

            if not cls.key_contained(key, cls.__slots__):
                # it has something we don't - probably not a match
                return False

        # then, ensure that all required fields are present
        for key in cls.required_fields:
            if not cls.key_contained(key, dct):
                # it doesn't have everything we need - probably not a match
                return False

        return True

    def _parse_spec_extensions(self):
        """
        Examines the keys of this Object's raw_element and collects any `Specification
        Extensions`_ into the extensions attribute of this Object.

        .. _Specification Extensions: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#specificationExtensions
        """
        for k, v in self.raw_element.items():
            if k.startswith("x-"):
                self.extensions[k[2:]] = v
                self._accessed_members.append(k)

    @classmethod
    def get_object_type(cls, typename):
        """
        Introspects the subclasses of this class to decide which to return for
        object_type

        :param object_type: The name of a class that inherits from this class.
                            Must exactly equal type(Class).__name__
        :type object_type: str

        :returns: The Type associated with this name
        :raises ValueError: if no Type with that name was found
        """
        if not hasattr(cls, "_subclass_map"):
            # generate subclass map on first call
            setattr(cls, "_subclass_map", {t.__name__: t for t in cls.__subclasses__()})

        # TODO - why?
        if typename not in cls._subclass_map:  # pylint: disable=no-member
            raise ValueError("ObjectBase has no subclass {}".format(typename))

        return cls._subclass_map[typename]  # pylint: disable=no-member

    def get_path(self):
        """
        Get the full path for this element in the spec

        :returns: The path in the spec for this element
        :rtype: str
        """
        return ".".join(self.path)

    def parse_list(self, raw_list, object_types, field=None):
        """
        Given a list of Objects, iterates over the list and creates the relevant
        Objects, returning the resulting list.

        :param raw_list: The list to parse
        :type raw_list: list[dict]
        :param object_types: A list of subclass names to attempt to parse the
                             objects to.  The list does not need to consist of
                             only one of these types.
        :type object_type: list[str]
        :param field: The field to append to self.get_path() when determining path
                      for created objects.
        :type field: str

        :returns: A list of parsed objects
        :rtype: list[object_type]
        """
        if raw_list is None:
            return None

        if not isinstance(object_types, list):
            object_types = [object_types]

        real_path = self.path[:]
        if field:
            real_path += [field]

        python_types = [ObjectBase.get_object_type(t) if isinstance(t, str) else t for t in object_types]

        result = []
        for i, cur in enumerate(raw_list):
            found_type = False

            for cur_type in python_types:
                if issubclass(cur_type, ObjectBase) and cur_type.can_parse(cur):
                    result.append(cur_type(real_path + [str(i)], cur, self._root))
                    found_type = True
                    continue
                elif isinstance(cur, cur_type):
                    result.append(cur)
                    found_type = True
                    continue

            if not found_type:
                raise SpecError(
                    "Could not parse {}.{}, expected to be one of [{}]".format(".".join(real_path), i, object_types),
                    path=self.path,
                    element=self,
                )

        return result

    def _resolve_references(self):
        """
        Resolves all reference objects below this object and notes their original
        value was a reference.
        """
        # don't circular import
        reference_type = ObjectBase.get_object_type("Reference")

        for slot in self.__slots__:
            if slot.startswith("_"):
                # don't parse private members
                continue
            value = getattr(self, slot)

            if isinstance(value, reference_type):
                # we found a reference - attempt to resolve it
                reference_path = value.ref
                if not reference_path.startswith("#/"):
                    raise ReferenceResolutionError(
                        "Invalid reference path {}".format(reference_path), path=self.path, element=self
                    )

                reference_path = reference_path.split("/")[1:]

                try:
                    resolved_value = self._root.resolve_path(reference_path)
                except ReferenceResolutionError as e:
                    # add metadata to the error
                    e.path = self.path
                    e.element = self
                    raise

                # FIXME - will break if multiple things reference the same
                # node
                resolved_value._original_ref = value

                # resolved
                setattr(self, slot, resolved_value)
            elif issubclass(type(value), ObjectBase) or isinstance(value, Map):
                # otherwise, continue resolving down the tree
                value._resolve_references()
            elif isinstance(value, list):
                # if it's a list, resolve its item's references
                resolved_list = []
                for item in value:
                    if isinstance(item, reference_type):
                        # TODO - this is duplicated code
                        reference_path = item.ref.split("/")[1:]

                        try:
                            resolved_value = self._root.resolve_path(reference_path)
                        except ReferenceResolutionError as e:
                            # add metadata to the error
                            e.path = self.path
                            e.element = self
                            raise

                        resolved_value._original_ref = value
                        resolved_list.append(resolved_value)
                    else:
                        if issubclass(type(value), ObjectBase) or isinstance(value, Map):
                            item._resolve_references()
                        resolved_list.append(item)

                setattr(self, slot, resolved_list)

    def _resolve_allOfs(self):
        """
        Walks object tree calling _resolve_allOf on each type.

        Types can override this to handle allOf handling themselves.  Types that
        do so should call the parent class' _resolve_allOf when they do
        """
        for slot in self.__slots__:
            if slot.startswith("_"):
                # no need to handle private members
                continue

            value = getattr(self, slot)

            if issubclass(type(value), ObjectBase):
                value._resolve_allOfs()
            elif issubclass(type(value), Map):
                for _, c in value.items():
                    c._resolve_allOfs()
            elif isinstance(value, list):
                for c in value:
                    if issubclass(type(c), ObjectBase) or issubclass(type(c), Map):
                        c._resolve_allOfs()


class Map(dict):
    """
    The Map object wraps a python dict and parses its values into the chosen
    type or types.
    """

    __slots__ = ["dct", "path", "raw_element", "_root"]

    def __init__(self, path, raw_element, object_types, root):
        """
        Creates a dict containing the parsed objects from the raw element

        :param path: The path to this Map in the spec.
        :type path: list
        :param raw_element: The raw spec data for this map.  The keys must all
                            be strings.
        :type raw_element: dict
        :param object_types: A list of strings accepted by
                             :any:`ObjectBase.get_object_type`, or the python
                             types to parse.
        :type object_types: list[str or Type]
        """
        self.path = path
        self.raw_element = raw_element
        self._root = root

        python_types = []
        dct = {}

        for t in object_types:
            if isinstance(t, str):
                python_types.append(ObjectBase.get_object_type(t))
            else:
                python_types.append(t)

        for k, v in self.raw_element.items():
            found_type = False

            for t in python_types:
                if issubclass(t, ObjectBase) and t.can_parse(v):
                    dct[k] = t(path + [k], v, self._root)
                    found_type = True
                elif isinstance(v, t):
                    dct[k] = v
                    found_type = True

            if not found_type:
                raise_on_unknown_type(self, k, object_types, v)

        self.update(dct)

    def _resolve_references(self):
        """
        This has been added to allow propagation of reference resolution as defined
        in :any:`ObjectBase._resolve_references`.  This implementation simply
        calls the same on all values in this Map.
        """
        reference_type = ObjectBase.get_object_type("Reference")

        for key, value in self.items():
            if isinstance(value, reference_type):
                # TODO - this is repeated code
                # we found a reference - attempt to resolve it
                reference_path = value.ref
                if not reference_path.startswith("#/"):
                    raise ReferenceResolutionError(
                        "Invalid reference path {}".format(reference_path), path=self.path, element=self
                    )

                reference_path = reference_path.split("/")[1:]

                try:
                    resolved_value = self._root.resolve_path(reference_path)
                except ReferenceResolutionError as e:
                    # add metadata to the error
                    e.path = self.path
                    e.element = self
                    raise

                # FIXME - will break if multiple things reference the same
                # node
                resolved_value._original_ref = value

                # resolved
                self[key] = resolved_value
            else:
                value._resolve_references()

    def get_path(self):
        """
        Get the full path for this element in the spec

        :returns: The path in the spec for this element
        :rtype: str
        """
        return ".".join(self.path)



from .errors import SpecError
from .object_base import ObjectBase, Map


class SecurityScheme(ObjectBase):
    """
    A `Security Scheme`_ defines a security scheme that can be used by the operations.

    .. _Security Scheme: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#securitySchemeObject
    """

    __slots__ = ["type", "description", "name", "in", "in_", "scheme", "bearerFormat", "flows", "openIdConnectUrl"]
    required_fields = ["type"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.bearerFormat = self._get("bearerFormat", [str])
        self.description = self._get("description", [str])
        self.flows = self._get("flows", dict)  # ['OAuthFlows']) TODO
        self.in_ = self._get("in", str)
        self.name = self._get("name", [str])
        self.openIdConnectUrl = self._get("openIdConnectUrl", [str])
        self.scheme = self._get("scheme", [str])
        self.type = self._get("type", [str])



from .object_base import ObjectBase


class Example(ObjectBase):
    """
    A `Example Object`_ holds a reusable set of different aspects of the OAS
    spec.

    .. _Example Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#exampleObject
    """

    __slots__ = ["summary", "description", "value", "externalValue"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.summary = self._get("summary", str)
        self.description = self._get("description", str)
        self.value = self._get("value", ["Reference", dict, str])  # 'any' type
        self.externalValue = self._get("externalValue", str)



from .object_base import ObjectBase


class Tag(ObjectBase):
    """
    A `Tag Object`_ holds a reusable set of different aspects of the OAS
    spec.

    .. _Tag Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#tagObject
    """

    __slots__ = ["name", "description", "externalDocs"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.name = self._get("name", str)
        self.description = self._get("description", str)
        self.externalDocs = self._get("externalDocs", str)



class SpecError(ValueError):
    """
    This error class is used when an invalid format is found while parsing an
    object in the spec.
    """

    def __init__(self, message, path=None, element=None):
        self.element = element
        self.message = message
        self.path = path


class ReferenceResolutionError(SpecError):
    """
    This error class is used when resolving a reference fails, usually because
    of a malformed path in the reference.
    """


class ModelError(ValueError):
    """The data supplied to the Model mismatches the models attributes"""



from .object_base import ObjectBase


class Info(ObjectBase):
    """
    An OpenAPI Info object, as defined in `the spec`_.

    .. _the spec: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#infoObject
    """

    __slots__ = ["title", "description", "termsOfService", "contact", "license", "version"]
    required_fields = ["title", "version"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.contact = self._get("contact", "Contact")
        self.description = self._get("description", str)
        self.license = self._get("license", "License")
        self.termsOfService = self._get("termsOfService", str)
        self.title = self._get("title", str)
        self.version = self._get("version", str)


class Contact(ObjectBase):
    """
    Contact object belonging to an Info object, as described `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#contactObject
    """

    __slots__ = ["name", "url", "email"]
    required_fields = ["name", "url", "email"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.email = self._get("email", str)
        self.name = self._get("name", str)
        self.url = self._get("url", str)


class License(ObjectBase):
    """
    License object belonging to an Info object, as described `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#license-object
    """

    __slots__ = ["name", "url"]
    required_fields = ["name"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.name = self._get("name", str)
        self.url = self._get("url", str)



from .errors import SpecError, ModelError
from .general import Reference  # need this for Model below
from .object_base import ObjectBase, Map

TYPE_LOOKUP = {
    "array": list,
    "integer": int,
    "object": dict,
    "string": str,
    "boolean": bool,
}


class Schema(ObjectBase):
    """
    The `Schema Object`_ allows the definition of input and output data types.

    .. _Schema Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#schemaObject
    """

    __slots__ = [
        "title",
        "multipleOf",
        "maximum",
        "exclusiveMaximum",
        "minimum",
        "exclusiveMinimum",
        "maxLength",
        "minLength",
        "pattern",
        "maxItems",
        "minItems",
        "uniqueItems",
        "maxProperties",
        "minProperties",
        "required",
        "enum",
        "type",
        "allOf",
        "oneOf",
        "anyOf",
        "not",
        "items",
        "properties",
        "additionalProperties",
        "description",
        "format",
        "default",
        "nullable",
        "discriminator",
        "readOnly",
        "writeOnly",
        "xml",
        "externalDocs",
        "example",
        "deprecated",
        "contentEncoding",
        "contentMediaType",
        "contentSchema",
        "_model_type",
        "_request_model_type",
        "_resolved_allOfs",
    ]
    required_fields = []

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.title = self._get("title", str)
        self.maximum = self._get("maximum", [int, float])
        self.minimum = self._get("minimum", [int, float])
        self.maxLength = self._get("maxLength", int)
        self.minLength = self._get("minLength", int)
        self.pattern = self._get("pattern", str)
        self.maxItems = self._get("maxItems", int)
        self.minItems = self._get("minItmes", int)
        self.required = self._get("required", list)
        self.enum = self._get("enum", list)
        self.type = self._get("type", str)
        self.allOf = self._get("allOf", ["Schema", "Reference"], is_list=True)
        self.oneOf = self._get("oneOf", list)
        self.anyOf = self._get("anyOf", list)
        self.items = self._get("items", ["Schema", "Reference"])
        self.properties = self._get("properties", ["Schema", "Reference"], is_map=True)
        self.additionalProperties = self._get("additionalProperties", [bool, dict])
        self.description = self._get("description", str)
        self.format = self._get("format", str)
        self.default = self._get("default", TYPE_LOOKUP.get(self.type, str))  # TODO - str as a default?
        self.nullable = self._get("nullable", bool)
        self.discriminator = self._get("discriminator", dict)  # 'Discriminator'
        self.readOnly = self._get("readOnly", bool)
        self.writeOnly = self._get("writeOnly", bool)
        self.xml = self._get("xml", dict)  # 'XML'
        self.externalDocs = self._get("externalDocs", dict)  # 'ExternalDocs'
        self.deprecated = self._get("deprecated", bool)
        self.example = self._get("example", "*")
        self.contentEncoding = self._get("contentEncoding", str)
        self.contentMediaType = self._get("contentMediaType", str)
        self.contentSchema = self._get("contentSchema", str)

        # TODO - Implement the following properties:
        # self.multipleOf
        # self.not
        # self.uniqueItems
        # self.maxProperties
        # self.minProperties
        # self.exclusiveMinimum
        # self.exclusiveMaximum

        self._resolved_allOfs = False

        if self.type == "array" and self.items is None:
            raise SpecError('{}: items is required when type is "array"'.format(self.get_path()))

    def get_type(self):
        """
        Returns the Type that this schema represents.  This Type is created once
        per Schema and cached so that all instances of the same schema are the
        same Type.  For example::

           object1 = example_schema.model({"some":"json"})
           object2 = example_schema.model({"other":"json"})

           isinstance(object1, example._schema.get_type()) # true
           type(object1) == type(object2) # true
        """
        # this is defined in ObjectBase.__init__ as all slots are
        if self._model_type is None:  # pylint: disable=access-member-before-definition
            type_name = self.title or self.path[-1]
            self._model_type = type(
                type_name,
                (Model,),
                {"__slots__": self.properties.keys()},  # pylint: disable=attribute-defined-outside-init
            )

        return self._model_type

    def model(self, data):
        """
        Generates a model representing this schema from the given data.

        :param data: The data to create the model from.  Should match this schema.
        :type data: dict

        :returns: A new :any:`Model` created in this Schema's type from the data.
        :rtype: self.get_type()
        """
        if self.properties is None and self.type in ("string", "number"):  # more simple types
            # if this schema represents a simple type, simply return the data
            # TODO - perhaps assert that the type of data matches the type we
            # expected
            return data
        elif self.type == "array":
            return [self.items.get_type()(i, self.items) for i in data]
        else:
            return self.get_type()(data, self)

    def get_request_type(self):
        """
        Similar to :any:`get_type`, but the resulting type does not accept readOnly
        fields
        """
        # this is defined in ObjectBase.__init__ as all slots are
        if self._request_model_type is None:  # pylint: disable=access-member-before-definition
            type_name = self.title or self.path[-1]
            self._request_model_type = type(
                type_name + "Request",
                (Model,),
                {  # pylint: disable=attribute-defined-outside-init
                    "__slots__": [k for k, v in self.properties.items() if not v.readOnly]
                },
            )

        return self._request_model_type

    def request_model(self, **kwargs):
        """
        Converts the kwargs passed into a model of writeable fields of this
        schema
        """
        # TODO - this doesn't get nested schemas
        return self.get_request_type()(kwargs, self)

    def _resolve_allOfs(self):
        """
        Handles merging properties for allOfs
        """
        if self._resolved_allOfs:
            return

        self._resolved_allOfs = True

        if self.allOf:
            for c in self.allOf:
                if isinstance(c, Schema):
                    self._merge(c)

    def _merge(self, other):
        """
        Merges ``other`` into this schema, preferring to use the values in ``other``
        """
        for slot in self.__slots__:
            if slot.startswith("_"):
                # skip private members
                continue

            my_value = getattr(self, slot)
            other_value = getattr(other, slot)

            if other_value:
                # we got a value to merge
                if isinstance(other_value, Schema):
                    # if it's another schema, merge them
                    if my_value is not None:
                        my_value._merge(other_value)
                    else:
                        setattr(self, slot, other_value)
                elif isinstance(other_value, list):
                    # we got a list, combine them
                    if my_value is None:
                        my_value = []
                    setattr(self, slot, my_value + other_value)
                elif isinstance(other_value, dict) or isinstance(other_value, Map):
                    if my_value:
                        for k, v in my_value.items():
                            if k in other_value:
                                if isinstance(v, Schema):
                                    v._merge(other_value[k])
                                    continue
                                else:
                                    my_value[k] = other_value[k]
                        for ok, ov in other_value.items():
                            if ok not in my_value:
                                my_value[ok] = ov
                    else:
                        setattr(self, slot, other_value)
                else:
                    setattr(self, slot, other_value)


class Model:
    """
    A Model is a representation of a Schema as a request or response.  Models
    are generated from Schema objects by called :any:`Schema.model` with the
    contents of a response.
    """

    __slots__ = ["_raw_data", "_schema"]

    def __init__(self, data, schema):
        """
        Creates a new Model from data.  This should never be called directly,
        but instead should be called through :any:`Schema.model` to generate a
        Model from a defined Schema.

        :param data: The data to create this Model with
        :type data: dict
        """
        self._raw_data = data
        self._schema = schema

        for s in self.__slots__:
            # initialize all slots to None
            setattr(self, s, None)

        keys = set(data.keys()) - frozenset(self.__slots__)
        if keys:
            raise ModelError("Schema {} got unexpected attribute keys {}".format(self.__class__.__name__, keys))

        # collect the data into this model
        for k, v in data.items():
            prop = schema.properties[k]

            if prop.type == "array":
                # handle arrays
                item_schema = prop.items
                setattr(self, k, [item_schema.model(c) for c in v])
            elif prop.type == "object":
                # handle nested objects
                object_schema = prop
                setattr(self, k, object_schema.model(v))
            else:
                setattr(self, k, v)

    def __repr__(self):
        """
        A generic representation of this model
        """
        return str(dict(self))

    def __iter__(self):
        for s in self.__slots__:
            if s.startswith("_"):
                continue
            yield s, getattr(self, s)
        return



import requests

from .object_base import ObjectBase, Map
from .errors import ReferenceResolutionError, SpecError


class OpenAPI(ObjectBase):
    """
    This class represents the root of the OpenAPI schema document, as defined
    in `the spec`_

    .. _the spec: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#openapi-object
    """

    __slots__ = [
        "openapi",
        "info",
        "servers",
        "paths",
        "components",
        "security",
        "tags",
        "externalDocs",
        "_operation_map",
        "_security",
        "validation_mode",
        "_spec_errors",
        "_ssl_verify",
        "_session",
    ]
    required_fields = ["openapi", "info", "paths"]

    def __init__(
        self, raw_document, validate=False, ssl_verify=None, use_session=False, session_factory=requests.Session
    ):
        """
        Creates a new OpenAPI document from a loaded spec file.  This is
        overridden here because we need to specify the path in the parent
        class' constructor.

        :param raw_document: The raw OpenAPI file loaded into python
        :type raw_document: dct
        :param validate: If True, don't fail on errors, but instead capture all
                         errors, continuing along the spec as best as possible,
                         and make them available when parsing is complete.
        :type validate: bool
        :param ssl_verify: Decide if to use ssl verification to the requests or not,
                           in case an str is passed, will be used as the CA.
        :type ssl_verify: bool, str, None
        :param use_session: Should we use a consistant session between API calls
        :type use_session: bool
        """
        # do this first so super().__init__ can see it
        self.validation_mode = validate

        if validate:
            self._spec_errors = []

        # as the document root, we have no path
        super(OpenAPI, self).__init__([], raw_document, self)

        self._security = {}

        self._ssl_verify = ssl_verify

        self._session = None
        if use_session:
            self._session = session_factory()

    # public methods
    def authenticte(self, security_scheme, value):
        """
        Authenticates all subsequent requests with the given arguments.

        TODO - this should support more than just HTTP Auth
        """

        # authentication is optional and can be disabled
        if security_scheme is None:
            self._security = None
            return

        if security_scheme not in self.components.securitySchemes:
            raise ValueError("{} does not accept security scheme {}".format(self.info.title, security_scheme))

        self._security = {security_scheme: value}

    authenticate = authenticte

    def resolve_path(self, path):
        """
        Given a $ref path, follows the document tree and returns the given attribute.

        :param path: The path down the spec tree to follow
        :type path: list[str]

        :returns: The node requested
        :rtype: ObjectBase
        :raises ValueError: if the given path is not valid
        """
        node = self

        for part in path:
            if isinstance(node, Map):
                if part not in node:  # pylint: disable=unsupported-membership-test
                    err_msg = "Invalid path {} in Reference".format(path)
                    raise ReferenceResolutionError(err_msg)
                node = node.get(part)
            else:
                if not hasattr(node, part):
                    err_msg = "Invalid path {} in Reference".format(path)
                    raise ReferenceResolutionError(err_msg)
                node = getattr(node, part)

        return node

    def log_spec_error(self, error):
        """
        In Validation Mode, this method is used when parsing a spec to record an
        error that was encountered, for later reporting.  This should not be used
        outside of Validation Mode.

        :param error: The error encountered.
        :type error: SpecError
        """
        if not self.validation_mode:
            raise RuntimeError("This client is not in Validation Mode, cannot " "record errors!")
        self._spec_errors.append(error)

    def errors(self):
        """
        In Validation Mode, returns all errors encountered from parsing a spec.
        This should not be called if not in Validation Mode.

        :returns: The errors encountered during the parsing of this spec.
        :rtype: list[SpecError]
        """
        if not self.validation_mode:
            raise RuntimeError("This client is not in Validation Mode, cannot " "return errors!")
        return self._spec_errors

    # private methods
    def _register_operation(self, operation_id, operation):
        """
        Adds an Operation to this spec's _operation_map, raising an error if the
        OperationId has already been registered.

        :param operation_id: The operation ID to register
        :type operation_id: str
        :param operation: The operation to register
        :type operation: Operation
        """
        if operation_id in self._operation_map:
            raise SpecError("Duplicate operationId {}".format(operation_id), path=operation.path)
        self._operation_map[operation_id] = operation

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self._operation_map = {}

        self.components = self._get("components", ["Components"])
        self.externalDocs = self._get("externalDocs", dict)
        self.info = self._get("info", "Info")
        self.openapi = self._get("openapi", str)
        self.paths = self._get("paths", ["Path"], is_map=True)
        self.security = self._get("security", ["SecurityRequirement"], is_list=True)
        self.servers = self._get("servers", ["Server"], is_list=True)
        self.tags = self._get("tags", ["Tag"], is_list=True)

        # now that we've parsed _all_ the data, resolve all references
        self._resolve_references()
        self._resolve_allOfs()

    def _get_callable(self, operation):
        """
        A helper function to create OperationCallable objects for __getattribute__,
        pre-initialized with the required values from this object.

        :param operation: The Operation the callable should call
        :type operation: callable (Operation.request)

        :returns: The callable that executes this operation with this object's
                  configuration.
        :rtype: OperationCallable
        """
        base_url = self.servers[0].url

        return OperationCallable(operation, base_url, self._security, self._ssl_verify, self._session)

    def __getattribute__(self, attr):
        """
        Extended __getattribute__ function to allow resolving dynamic function
        names.  The purpose of this is to call syntax like this::

           spec = OpenAPI(raw_spec)
           spec.call_operationId()

        This method will intercept the dot notation above (spec.call_operationId)
        and look up the requested operation, returning a callable object that
        will then immediately be called by the parenthesis.

        :param attr: The attribute we're retrieving
        :type attr: str

        :returns: The attribute requested
        :rtype: any
        :raises AttributeError: if the requested attribute does not exist
        """
        if attr.startswith("call_"):
            _, operationId = attr.split("_", 1)
            if operationId in self._operation_map:
                return self._get_callable(self._operation_map[operationId].request)
            else:
                raise AttributeError("{} has no operation {}".format(self.info.title, operationId))

        return object.__getattribute__(self, attr)


class OperationCallable:
    """
    This class is returned by instances of the OpenAPI class when members
    formatted like call_operationId are accessed, and a valid Operation is
    found, and allows calling the operation directly from the OpenAPI object
    with the configured values included.  This class is not intended to be used
    directly.
    """

    def __init__(self, operation, base_url, security, ssl_verify, session):
        self.operation = operation
        self.base_url = base_url
        self.security = security
        self.ssl_verify = ssl_verify
        self.session = session

    def __call__(self, *args, **kwargs):
        if self.ssl_verify is not None:
            kwargs["verify"] = self.ssl_verify
        if self.session:
            kwargs["session"] = self.session
        return self.operation(self.base_url, *args, security=self.security, **kwargs)



from .object_base import ObjectBase


class Server(ObjectBase):
    """
    The Server object, as described `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#serverObject
    """

    __slots__ = ["url", "description", "variables"]
    required_fields = ["url"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.description = self._get("description", str)
        self.url = self._get("url", str)
        self.variables = self._get("variables", ["ServerVariable"], is_map=True)


class ServerVariable(ObjectBase):
    """
    A ServerVariable object as defined `here`_.

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#server-variable-object
    """

    __slots__ = ["enum", "default", "description"]
    required_fields = ["default"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.default = self._get("default", str)
        self.description = self._get("description", str)
        self.enum = self._get("enum", [str], is_list=True)



import sys
import yaml

from .openapi import OpenAPI


def main():
    specfile = sys.argv[1]

    with open(specfile) as f:
        spec = yaml.safe_load(f.read())

    o = OpenAPI(spec, validate=True)

    errors = o.errors()

    if errors:
        # print errors
        for e in errors:
            print("{}: {}".format(".".join(e.path), e.message[:300]))
        print()
        print("{} errors".format(len(errors)))
        sys.exit(1)  # exit with error status
    else:
        print("OK")


if __name__ == "__main__":
    main()



import json
import re
import requests

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from .errors import SpecError
from .object_base import ObjectBase
from .schemas import Model


def _validate_parameters(instance):
    """
    Ensures that all parameters for this path are valid
    """
    allowed_path_parameters = re.findall(r"{([a-zA-Z0-9\-\._~]+)}", instance.path[1])

    for c in instance.parameters:
        if c.in_ == "path":
            if c.name not in allowed_path_parameters:
                raise SpecError("Parameter name not found in path: {}".format(c.name), path=instance.path)


class Path(ObjectBase):
    """
    A Path object, as defined `here`_.  Path objects represent URL paths that
    may be accessed by appending them to a Server

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#paths-object
    """

    __slots__ = [
        "summary",
        "description",
        "get",
        "put",
        "post",
        "delete",
        "options",
        "head",
        "patch",
        "trace",
        "servers",
        "parameters",
    ]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        # TODO - handle possible $ref
        self.delete = self._get("delete", "Operation")
        self.description = self._get("description", str)
        self.get = self._get("get", "Operation")
        self.head = self._get("head", "Operation")
        self.options = self._get("options", "Operation")
        self.parameters = self._get("parameters", ["Parameter", "Reference"], is_list=True)
        self.patch = self._get("patch", "Operation")
        self.post = self._get("post", "Operation")
        self.put = self._get("put", "Operation")
        self.servers = self._get("servers", ["Server"], is_list=True)
        self.summary = self._get("summary", str)
        self.trace = self._get("trace", "Operation")

        if self.parameters is None:
            # this will be iterated over later
            self.parameters = []

    def _resolve_references(self):
        """
        Overloaded _resolve_references to allow us to verify parameters after
        we've got all references settled.
        """
        super(self.__class__, self)._resolve_references()

        # this will raise if parameters are invalid
        _validate_parameters(self)


class Parameter(ObjectBase):
    """
    A `Parameter Object`_ defines a single operation parameter.

    .. _Parameter Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#parameterObject
    """

    __slots__ = [
        "name",
        "in",
        "in_",
        "description",
        "required",
        "deprecated",
        "allowEmptyValue",
        "style",
        "explode",
        "allowReserved",
        "schema",
        "example",
        "examples",
    ]
    required_fields = ["name", "in"]

    def _parse_data(self):
        self.deprecated = self._get("deprecated", bool)
        self.description = self._get("description", str)
        self.example = self._get("example", str)
        self.examples = self._get("examples", dict)  # Map[str: ['Example','Reference']]
        self.explode = self._get("explode", bool)
        self.in_ = self._get("in", str)  # TODO must be one of ["query","header","path","cookie"]
        self.name = self._get("name", str)
        self.required = self._get("required", bool)
        self.schema = self._get("schema", ["Schema", "Reference"])
        self.style = self._get("style", str)

        # allow empty or reserved values in Parameter data
        self.allowEmptyValue = self._get("allowEmptyValue", bool)
        self.allowReserved = self._get("allowReserved", bool)

        # required is required and must be True if this parameter is in the path
        if self.in_ == "path" and self.required is not True:
            err_msg = "Parameter {} must be required since it is in the path"
            raise SpecError(err_msg.format(self.get_path()), path=self.path)


class Operation(ObjectBase):
    """
    An Operation object as defined `here`_

    .. _here: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#operationObject
    """

    __slots__ = [
        "tags",
        "summary",
        "description",
        "externalDocs",
        "security",
        "operationId",
        "parameters",
        "requestBody",
        "responses",
        "callbacks",
        "deprecated",
        "servers",
        "_session",
        "_request",
    ]
    required_fields = ["responses"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        raw_servers = self._get("servers", list)
        self.deprecated = self._get("deprecated", bool)
        self.description = self._get("description", str)
        self.externalDocs = self._get("externalDocs", "ExternalDocumentation")
        self.operationId = self._get("operationId", str)
        self.parameters = self._get("parameters", ["Parameter", "Reference"], is_list=True)
        self.requestBody = self._get("requestBody", ["RequestBody", "Reference"])
        self.responses = self._get("responses", ["Response", "Reference"], is_map=True)
        self.security = self._get("security", ["SecurityRequirement"], is_list=True)
        self.servers = self._get("servers", ["Server"], is_list=True)
        self.summary = self._get("summary", str)
        self.tags = self._get("tags", list)
        raw_servers = self._get("servers", list)
        # self.callbacks  = self._get('callbacks', dict) TODO

        # default parameters to an empty list for processing later
        if self.parameters is None:
            self.parameters = []

        # gather all operations into the spec object
        if self.operationId is not None:
            # TODO - how to store without an operationId?
            formatted_operation_id = self.operationId.replace(" ", "_")
            self._root._register_operation(formatted_operation_id, self)

        # TODO - maybe make this generic
        if self.security is None:
            self.security = self._root._get("security", ["SecurityRequirement"], is_list=True) or []

        # Store session object
        self._session = requests.Session()

        # Store request object
        self._request = requests.Request()

    def _resolve_references(self):
        """
        Overloaded _resolve_references to allow us to verify parameters after
        we've got all references settled.
        """
        super(self.__class__, self)._resolve_references()

        # this will raise if parameters are invalid
        _validate_parameters(self)

    def _request_handle_secschemes(self, security_requirement, value):
        ss = self._root.components.securitySchemes[security_requirement.name]

        if ss.type == "http" and ss.scheme == "basic":
            self._request.auth = requests.auth.HTTPBasicAuth(*value)

        if ss.type == "http" and ss.scheme == "digest":
            self._request.auth = requests.auth.HTTPDigestAuth(*value)

        if ss.type == "http" and ss.scheme == "bearer":
            header = ss.bearerFormat or "Bearer {}"
            self._request.headers["Authorization"] = header.format(value)

        if ss.type == "mutualTLS":
            # TLS Client certificates (mutualTLS)
            self._request.cert = value

        if ss.type == "apiKey":
            if ss.in_ == "query":
                # apiKey in query parameter
                self._request.params[ss.name] = value

            if ss.in_ == "header":
                # apiKey in query header data
                self._request.headers[ss.name] = value

            if ss.in_ == "cookie":
                self._request.cookies = {ss.name: value}

    def _request_handle_parameters(self, parameters={}):
        # Parameters
        path_parameters = {}
        accepted_parameters = {}
        p = self.parameters + self._root.paths[self.path[-2]].parameters

        for _ in list(p):
            # TODO - make this work with $refs - can operations be $refs?
            accepted_parameters.update({_.name: _})

        for name, spec in accepted_parameters.items():
            try:
                value = parameters[name]
            except KeyError:
                if spec.required and name not in parameters:
                    err_msg = "Required parameter {} not provided".format(name)
                    raise ValueError(err_msg)

                continue

            if spec.in_ == "path":
                # The string method `format` is incapable of partial updates,
                # as such we need to collect all the path parameters before
                # applying them to the format string.
                path_parameters[name] = value

            if spec.in_ == "query":
                self._request.params[name] = value

            if spec.in_ == "header":
                self._request.headers[name] = value

            if spec.in_ == "cookie":
                self._request.cookies[name] = value

        self._request.url = self._request.url.format(**path_parameters)

    def _request_handle_body(self, data):
        if "application/json" in self.requestBody.content:
            if isinstance(data, dict) or isinstance(data, list):
                body = json.dumps(data)

            if issubclass(type(data), Model):
                # serialize models as dicts
                converter = lambda c: dict(c)
                data_dict = {k: v for k, v in data if v is not None}

                body = json.dumps(data_dict, default=converter)

            self._request.data = body
            self._request.headers["Content-Type"] = "application/json"
        else:
            raise NotImplementedError()

    def request(self, base_url, security={}, data=None, parameters={}, verify=True, session=None, raw_response=False):
        """
        Sends an HTTP request as described by this Path

        :param base_url: The URL to append this operation's path to when making
                         the call.
        :type base_url: str
        :param security: The security scheme to use, and the values it needs to
                         process successfully.
        :type security: dict{str: str}
        :param data: The request body to send.
        :type data: any, should match content/type
        :param parameters: The parameters used to create the path
        :type parameters: dict{str: str}
        :param verify: Should we do an ssl verification on the request or not,
                       In case str was provided, will use that as the CA.
        :type verify: bool/str
        :param session: a persistent request session
        :type session: None, requests.Session
        :param raw_response: If true, return the raw response instead of validating
                             and exterpolating it.
        :type raw_response: bool
        """
        # Set request method (e.g. 'GET')
        self._request = requests.Request(self.path[-1])

        # Set self._request.url to base_url w/ path
        self._request.url = base_url + self.path[-2]

        if security and self.security:
            security_requirement = None
            for scheme, value in security.items():
                security_requirement = None
                for r in self.security:
                    if r.name == scheme:
                        security_requirement = r
                        self._request_handle_secschemes(r, value)

            if security_requirement is None:
                err_msg = """No security requirement satisfied (accepts {}) \
                          """.format(
                    ", ".join(self.security.keys())
                )
                raise ValueError(err_msg)

        if self.requestBody:
            if self.requestBody.required and data is None:
                err_msg = "Request Body is required but none was provided."
                raise ValueError(err_msg)

            self._request_handle_body(data)

        self._request_handle_parameters(parameters)

        if session is None:
            session = self._session

        # send the prepared request
        result = session.send(self._request.prepare())

        # spec enforces these are strings
        status_code = str(result.status_code)

        # find the response model in spec we received
        expected_response = None
        if status_code in self.responses:
            expected_response = self.responses[status_code]
        elif "default" in self.responses:
            expected_response = self.responses["default"]

        if expected_response is None:
            # TODO - custom exception class that has the response object in it
            err_msg = """Unexpected response {} from {} (expected one of {}, \
                         no default is defined"""
            err_var = result.status_code, self.operationId, ",".join(self.responses.keys())

            raise RuntimeError(err_msg.format(*err_var))

        # if we got back a valid response code (or there was a default) and no
        # response content was expected, return None
        if expected_response.content is None:
            return

        content_type = result.headers["Content-Type"]
        expected_media = expected_response.content.get(content_type, None)

        if expected_media is None and "/" in content_type:
            # accept media type ranges in the spec. the most specific matching
            # type should always be chosen, but if we do not have a match here
            # a generic range should be accepted if one if provided
            # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#response-object

            generic_type = content_type.split("/")[0] + "/*"
            expected_media = expected_response.content.get(generic_type, None)

        if expected_media is None:
            err_msg = """Unexpected Content-Type {} returned for operation {} \
                         (expected one of {})"""
            err_var = result.headers["Content-Type"], self.operationId, ",".join(expected_response.content.keys())

            raise RuntimeError(err_msg.format(*err_var))

        response_data = None

        if content_type.lower() == "application/json":
            return expected_media.schema.model(result.json())
        else:
            raise NotImplementedError()


class SecurityRequirement(ObjectBase):
    """
    A `SecurityRequirement`_ object describes security schemes for API access.

    .. _SecurityRequirement: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#securityRequirementObject
    """

    ___slots__ = ["name", "types"]
    required_fields = []

    def _parse_data(self):
        """ """
        # usually these only ever have one key
        if len(self.raw_element.keys()) == 1:
            self.name = [c for c in self.raw_element.keys()][0]
            self.types = self._get(self.name, str, is_list=True)
        elif len(self.raw_element.keys()) == 0:
            # optional
            self.name = self.types = None

    @classmethod
    def can_parse(cls, dct):
        """
        This needs to ignore can_parse since the objects it's parsing are not
        regular - they must always have only one key though or be empty for Optional Security Requirements
        """
        return len(dct.keys()) == 1 and isinstance([c for c in dct.values()][0], list) or len(dct.keys()) == 0

    def __getstate__(self):
        return {self.name: self.types}


class RequestBody(ObjectBase):
    """
    A `RequestBody`_ object describes a single request body.

    .. _RequestBody: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#requestBodyObject
    """

    __slots__ = ["description", "content", "required"]
    required_fields = ["content"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.description = self._get("description", str)
        self.content = self._get("content", ["MediaType"], is_map=True)
        raw_content = self._get("content", dict)
        self.required = self._get("required", bool)


class MediaType(ObjectBase):
    """
    A `MediaType`_ object provides schema and examples for the media type identified
    by its key.  These are used in a RequestBody object.

    .. _MediaType: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#mediaTypeObject
    """

    __slots__ = ["schema", "example", "examples", "encoding"]
    required_fields = []

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.schema = self._get("schema", ["Schema", "Reference"])
        self.example = self._get("example", str)  # 'any' type
        self.examples = self._get("examples", ["Example", "Reference"], is_map=True)
        self.encoding = self._get("encoding", dict)  # Map['Encoding']


class Response(ObjectBase):
    """
    A `Response Object`_ describes a single response from an API Operation,
    including design-time, static links to operations based on the response.

    .. _Response Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#response-object
    """

    __slots__ = ["description", "headers", "content", "links"]
    required_fields = ["description"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.content = self._get("content", ["MediaType"], is_map=True)
        self.description = self._get("description", str)
        raw_content = self._get("content", dict)
        raw_headers = self._get("headers", dict)
        self.links = self._get("links", ["Link", "Reference"], is_map=True)


class Link(ObjectBase):
    """
    A `Link Object`_ describes a single Link from an API Operation Response to an API Operation Request

    .. _Link Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#linkObject
    """

    __slots__ = ["operationId", "operationRef", "description", "parameters", "requestBody", "server"]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.operationId = self._get("operationId", str)
        self.operationRef = self._get("operationRef", str)
        self.description = self._get("description", str)
        self.parameters = self._get("parameters", dict)
        self.requestBody = self._get("requestBody", dict)
        self.server = self._get("server", ["Server"])

        if self.operationId and self.operationRef:
            raise SpecError("operationId and operationRef are mutually exclusive, only one of them is allowed")
        if not (self.operationId or self.operationRef):
            raise SpecError("operationId and operationRef are mutually exclusive, one of them must be specified")



#!/usr/bin/env python3
from io import open
from setuptools import setup
from os import path


here = path.abspath(path.dirname(__file__))


# get the long description from the README.rst
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="openapi3",
    version="1.6.0",
    description="Client and Validator of OpenAPI 3 Specifications",
    long_description=long_description,
    author="dorthu",
    url="https://github.com/dorthu/openapi3",
    packages=["openapi3"],
    license="BSD 3-Clause License",
    install_requires=["PyYaml", "requests"],
    tests_require=["pytest", "pytest-asyncio", "uvloop", "hypercorn", "pydantic", "fastapi"],
)



from .object_base import ObjectBase


class Components(ObjectBase):
    """
    A `Components Object`_ holds a reusable set of different aspects of the OAS
    spec.

    .. _Components Object: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#componentsObject
    """

    __slots__ = [
        "schemas",
        "responses",
        "parameters",
        "examples",
        "headers",
        "requestBodies",
        "securitySchemes",
        "links",
        "callback",
    ]

    def _parse_data(self):
        """
        Implementation of :any:`ObjectBase._parse_data`
        """
        self.examples = self._get("examples", ["Example", "Reference"], is_map=True)
        self.parameters = self._get("parameters", ["Parameter", "Reference"], is_map=True)
        self.requestBodies = self._get("requestBody", ["RequestBody", "Reference"], is_map=True)
        self.responses = self._get("responses", ["Response", "Reference"], is_map=True)
        self.schemas = self._get("schemas", ["Schema", "Reference"], is_map=True)
        self.securitySchemes = self._get("securitySchemes", ["SecurityScheme", "Reference"], is_map=True)
        # self.headers       = self._get('headers', ['Header', 'Reference'], is_map=True)
        self.links = self._get("links", ["Link", "Reference"], is_map=True)
        # self.callbacks     = self._get('callbacks', ['Callback', 'Reference'], is_map=True)



from .openapi import OpenAPI

# these imports appear unused, but in fact load up the subclasses ObjectBase so
# that they may be referenced throughout the schema without issue
from . import info, servers, paths, general, schemas, components, security, tag, example
from .errors import SpecError, ReferenceResolutionError

__all__ = ["OpenAPI", "SpecError", "ReferenceResolutionError"]
